"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
var is_1 = __importDefault(require("@sindresorhus/is"));
var generate_id_1 = require("../../helpers/generate_id");
var Container = (function () {
    function Container(providers) {
        this.providers = providers;
        this.cache = {
            Request: new Map(),
            Singleton: new Map()
        };
    }
    Container.prototype.isBound = function (identifier) {
        return !!this.providers.find(function (x) { return x.provide === identifier; });
    };
    Container.prototype.get = function (identifier) {
        var _this = this;
        var id = generate_id_1.makeid();
        this.cache.Request.set(id, new Map());
        var build = this.build(identifier, id);
        if (is_1.default.promise(build)) {
            return build.then(function (instance) {
                _this.cache.Request.delete(id);
                return instance;
            });
        }
        else {
            this.cache.Request.delete(id);
            return build;
        }
    };
    Container.prototype.build = function (identifier, requestID) {
        var provider = this.providers.find(function (x) { return x.provide === identifier; });
        if (!provider) {
            throw new Error("No Provider for identifier " + this.printIdentifier(identifier));
        }
        if (provider.inject && provider.inject.length) {
            var injectScope = this.getScopeOfInjects(provider.inject);
            var scope = ScopePicker.pickList.apply(ScopePicker, [provider.scope].concat(injectScope));
            scope = provider.scope || "Singleton";
            switch (scope) {
                case "Singleton":
                    return this.buildAsSingleton(provider, requestID);
                case "Request":
                    return this.buildAsRequest(provider, requestID);
                case "Transient":
                    return this.buildAsTransient(provider, requestID);
            }
        }
        else {
            switch (provider.scope || "Singleton") {
                case "Singleton":
                    return this.buildAsSingleton(provider, requestID);
                case "Request":
                    return this.buildAsRequest(provider, requestID);
                case "Transient":
                    return this.buildAsTransient(provider, requestID);
                default:
                    throw new Error("Invalid scope " + provider.scope);
            }
        }
    };
    Container.prototype.buildAsSingleton = function (provider, requestID) {
        var _this = this;
        var cache = this.cache.Singleton;
        if (cache.has(provider.provide)) {
            return cache.get(provider.provide);
        }
        else {
            var dependencies = provider.inject.map(function (x) {
                return _this.build(x, requestID);
            });
            if (dependencies.some(is_1.default.promise)) {
                return Promise.all(dependencies.map(function (x) { return Promise.resolve(x); }))
                    .then(function (args) {
                    return provider.useFactory.apply(provider, args);
                })
                    .then(function (instance) {
                    cache.set(provider.provide, instance);
                    return instance;
                });
            }
            else {
                var instance = provider.useFactory.apply(provider, dependencies);
                cache.set(provider.provide, instance);
                return instance;
            }
        }
    };
    Container.prototype.buildAsRequest = function (provider, requestID) {
        var _this = this;
        var cache = this.cache.Request.get(requestID);
        if (cache.has(provider.provide)) {
            return cache.get(provider.provide);
        }
        else {
            var dependencies = (provider.inject || []).map(function (x) {
                return _this.build(x, requestID);
            });
            if (dependencies.some(is_1.default.promise)) {
                return Promise.all(dependencies.map(function (x) { return Promise.resolve(x); }))
                    .then(function (args) {
                    return provider.useFactory.apply(provider, args);
                })
                    .then(function (instance) {
                    cache.set(provider.provide, instance);
                    return instance;
                });
            }
            else {
                var instance = provider.useFactory.apply(provider, dependencies);
                cache.set(provider.provide, instance);
                return instance;
            }
        }
    };
    Container.prototype.buildAsTransient = function (provider, requestID) {
        var _this = this;
        var dependencies = (provider.inject || []).map(function (x) {
            return _this.build(x, requestID);
        });
        if (dependencies.some(is_1.default.promise)) {
            return Promise.all(dependencies.map(function (x) { return Promise.resolve(x); }))
                .then(function (args) {
                return provider.useFactory.apply(provider, args);
            })
                .then(function (instance) {
                return instance;
            });
        }
        else {
            var instance = provider.useFactory.apply(provider, dependencies);
            return instance;
        }
    };
    Container.prototype.printIdentifier = function (identifier) {
        return "" + identifier;
    };
    Container.prototype.getScopeOfInjects = function (injects) {
        var _this = this;
        var scopes = injects.map(function (inject) {
            var dependency = _this.providers.find(function (x) { return x.provide === inject; });
            if (!dependency) {
                throw new Error("Dependency not found for " + inject);
            }
            return dependency.scope;
        });
        return scopes;
    };
    Container.prototype.scopePicker = function (scope1, scope2, options) {
        if (options === void 0) { options = ["Transient", "Request", "Singleton"]; }
        return [scope1, scope2].sort(function (a, b) {
            return options.indexOf(a) - options.indexOf(b);
        })[0];
    };
    return Container;
}());
exports.Container = Container;
var ScopePicker = (function () {
    function ScopePicker() {
    }
    ScopePicker.pickList = function () {
        var _this = this;
        var scopes = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            scopes[_i] = arguments[_i];
        }
        return scopes.reduce(function (a, b) {
            return _this.pick(a, b);
        }, "Singleton");
    };
    ScopePicker.pick = function (scope1, scope2, options) {
        if (options === void 0) { options = ["Transient", "Request", "Singleton"]; }
        return [scope1, scope2].sort(function (a, b) {
            return options.indexOf(a) - options.indexOf(b);
        })[0];
    };
    return ScopePicker;
}());
exports.ScopePicker = ScopePicker;
